import torch
from einops import rearrange


class Factorization(torch.nn.Module):
    r"""
    Factorization model of the form $$X = UV^T$$

    where $X$ is the matrix to be factorized, $U$ and $V$ are the factor matrices.
    """

    def __init__(
        self,
        m_methods: int,
        n_examples: int,
        rank: int,
        ensemble_size: int,
        regularizer_weight: float = 0.00,
        drop_probability: float = 0.05,
    ) -> None:
        """
        Initialize the factorization model.

        Args:
            m_methods: Number of methods (i.e. rows of the matrix)
            n_examples: Number of examples (i.e. columns of the matrix)
            rank: Rank of the factorization
            ensemble_size: Number of factorization models in the ensemble
            regularizer_weight: Weight of the L2 regularizer
            drop_probability: Probability of dropping an entry in the matrix
        """
        super().__init__()
        self.register_buffer("U", torch.randn(ensemble_size, m_methods, rank))
        self.register_buffer("V", torch.randn(ensemble_size, n_examples, rank))
        self.register_buffer("L", regularizer_weight * torch.eye(rank))

        self.m_methods = m_methods
        self.n_examples = n_examples
        self.rank = rank
        self.ensemble_size = ensemble_size
        self.regularizer_weight = regularizer_weight
        self.drop_probability = drop_probability

    def forward(self) -> torch.Tensor:
        """
        Compute the approximation of the matrix using the factor matrices.
        """
        return torch.bmm(self.U, self.V.transpose(1, 2))

    def _als_step(
        self, data_matrix: torch.Tensor, fixed_matrix: torch.Tensor
    ) -> torch.Tensor:
        """
        Perform a single alternating least squares step.
        """
        non_zero_mask = (~torch.isnan(data_matrix)).float()
        y = fixed_matrix.unsqueeze(2)
        y_t = y.transpose(1, 2)
        A = (non_zero_mask.unsqueeze(2) * torch.bmm(y, y_t)).sum(0) + self.L
        b = (torch.nan_to_num(data_matrix * non_zero_mask) * y.squeeze(2)).sum(0)
        return torch.linalg.solve(A, b)

    def fit(self, X: torch.Tensor, iterations: int = 10) -> None:
        r"""
        Solve the following using alternating least squares

        $$\min_{U, V} \Big|\Big|O\odot\Big(UV^T - X\Big)\Big|\Big|_F^2$$

        where $O$ is the binary observation matrix (O_{ij} = 1 if X_{ij} is not
        nan else 0), $\odot$ is the Hadamard product, and $||\cdot||_F$ is the
        Frobenius norm.
        """
        X = X.unsqueeze(0).repeat(self.ensemble_size, 1, 1)
        if self.drop_probability > 0:
            mask = torch.rand_like(X) < self.drop_probability
            X[mask] = torch.nan
        X_u = rearrange(X, "e m n -> (e m) n 1")
        X_v = rearrange(X, "e m n -> (e n) m 1")
        vmap_als_step = torch.vmap(self._als_step, in_dims=(0, 0))
        for _ in range(iterations):
            self.V.data = (
                vmap_als_step(X_v, self.U.repeat(self.n_examples, 1, 1))
            ).reshape(self.ensemble_size, self.n_examples, self.rank)
            self.U.data = (
                vmap_als_step(X_u, self.V.repeat(self.m_methods, 1, 1))
            ).reshape(self.ensemble_size, self.m_methods, self.rank)

    def reset(self) -> None:
        """
        Reset the factor matrices to random values.
        """
        torch.nn.init.normal_(self.U)
        torch.nn.init.normal_(self.V)



import warnings

import torch

from banditeval.factorization import Factorization


def upper_confidence_bound_exploration(
    observed_matrix: torch.Tensor,
    a: int = 1,
    batch_size: int = 32,
    return_mus: bool = False,
):
    """Single step of Upper Confidence Bound Exploration"""
    m_methods, n_examples = observed_matrix.shape
    # Update bounds
    bounds = torch.full((m_methods,), torch.inf)
    mus = observed_matrix.nanmean(dim=1)
    counts = (~observed_matrix.isnan()).sum(1)
    mask = counts > 0
    bounds[mask] = mus[mask] + torch.sqrt(a / counts[mask])
    completely_sensed_mask = counts == n_examples
    bounds[completely_sensed_mask] = -torch.inf
    # Select batch
    if completely_sensed_mask.sum() == m_methods:
        batch = None
    else:
        best_method_index = torch.argmax(bounds)
        unobserved_column_indices = (
            observed_matrix[best_method_index].isnan().nonzero().flatten()
        )
        batch = torch.stack(
            [
                torch.full((batch_size,), best_method_index),
                unobserved_column_indices[
                    torch.randperm(unobserved_column_indices.size(0))[:batch_size]
                ],
            ]
        )
    if return_mus:
        return batch, mus
    return batch


def upper_confidence_bound_exploration_low_rank_factorization(
    observed_matrix: torch.Tensor,
    a: int = 1,
    batch_size: int = 32,
    rank: int = 1,
    ensemble_size: int = 64,
    warmup_percentage: float = 0.05,
    regularizer_weight: float = 0.1,
    drop_probability: float = 0.05,
    iterations: int = 10,
    eta: float = 5,
    device: str = "cpu",
    return_mus: bool = False,
):
    """Single step of Upper Confidence Bound Exploration with Low Rank Factorization"""
    if warmup_percentage < 0.05:
        warnings.warn(
            "Low warmup percentage can significantly degrade performance.", UserWarning
        )
    observed_matrix = observed_matrix.to(device)
    if (1 - observed_matrix.isnan().sum() / observed_matrix.numel()) < warmup_percentage:
        row_indices, column_indices = torch.where(observed_matrix.isnan())
        sample_indices = torch.randperm(row_indices.size(0))[:batch_size]
        batch = torch.stack([row_indices[sample_indices], column_indices[sample_indices]])
        if return_mus:
            warnings.warn("During warmup period scores are not computed.", UserWarning)
            return batch, None
        return batch
    m_methods, n_examples = observed_matrix.shape
    # Update bounds
    bounds = torch.full((m_methods,), torch.inf, device=device)
    factorization = Factorization(
        m_methods,
        n_examples,
        rank,
        ensemble_size,
        regularizer_weight=regularizer_weight,
        drop_probability=drop_probability,
    ).to(device)
    factorization.fit(observed_matrix, iterations=iterations)
    matrix_approximation = factorization()
    entry_mus = matrix_approximation.mean(0)
    entry_stds = matrix_approximation.std(0)
    entry_ucb = entry_mus + eta * entry_stds
    observed_mask = ~observed_matrix.isnan()
    entry_ucb[observed_mask] = observed_matrix[observed_mask]
    ucb = entry_ucb.mean(1)
    counts = observed_mask.sum(1)
    mask = counts > 0
    bounds[mask] = ucb[mask]
    completely_sensed_mask = counts == n_examples
    bounds[completely_sensed_mask] = -torch.inf
    # Select batch
    if completely_sensed_mask.sum() == m_methods:
        batch = None
    else:
        best_method_index = torch.argmax(bounds)
        best_method_stds = entry_stds[best_method_index]
        best_method_stds[observed_mask[best_method_index]] = -1
        batch = torch.stack(
            [
                torch.full((batch_size,), best_method_index, device=device),
                torch.topk(best_method_stds, batch_size, largest=True).indices,
            ]
        ).cpu()
    if return_mus:
        entry_mus[observed_mask] = observed_matrix[observed_mask]
        mus = entry_mus.mean(1).cpu()
        return batch, mus
    return batch



import pytest
import torch

from banditeval.factorization import Factorization


@pytest.fixture
def create_matrix():
    """ """

    def _create_matrix(m_methods, n_examples, rank):
        U = torch.rand(m_methods, rank)
        V = torch.rand(n_examples, rank)
        return torch.mm(U, V.T), U, V

    return _create_matrix


def test_single_factorization(create_matrix):
    """ """
    m_methods, n_examples = 10, 10
    ranks = [1, 3, 5]
    for rank in ranks:
        X, U, V = create_matrix(m_methods, n_examples, rank)
        factorization = Factorization(m_methods, n_examples, rank, 1, 0.1, 0.0)
        factorization.fit(X)
        assert torch.allclose(factorization().squeeze(0), X, atol=rank)
        assert torch.allclose(factorization.U.squeeze(0).abs(), U.abs(), atol=rank)
        assert torch.allclose(factorization.V.squeeze(0).abs(), V.abs(), atol=rank)


def test_ensembled_factorization(create_matrix):
    """ """
    m_methods, n_examples = 10, 10
    ranks = [1, 3, 5]
    for rank in ranks:
        X, U, V = create_matrix(m_methods, n_examples, rank)
        factorization = Factorization(m_methods, n_examples, rank, 32, 0.1, 0.05)
        factorization.fit(X)
        assert torch.allclose(factorization().mean(0), X, atol=rank)
        assert torch.allclose(factorization.U.mean(0).abs(), U.abs(), atol=rank)
        assert torch.allclose(factorization.V.mean(0).abs(), V.abs(), atol=rank)



import pytest
import torch

from banditeval.bandits import (
    upper_confidence_bound_exploration,
    upper_confidence_bound_exploration_low_rank_factorization,
)


@pytest.fixture
def create_obvious_winner_matrix():
    """ """

    def _create_matrix(m_methods, n_examples):
        X = torch.full((m_methods, n_examples), torch.nan)
        indices = torch.randperm(m_methods)
        winner, losers = indices[0:1], indices[1:]
        X[winner, : n_examples // 2] = 1
        X[losers, : n_examples // 2] = 0
        return X, winner.item()

    return _create_matrix


@pytest.fixture
def create_nearly_full_matrix():
    """ """

    def _create_matrix(m_methods, n_examples):
        X = torch.rand(m_methods, n_examples).round()
        method_index = torch.randint(m_methods, (1,))
        example_index = torch.randint(n_examples, (1,))
        X[method_index, example_index] = torch.nan
        return X, method_index.item()

    return _create_matrix


def test_upper_confidence_bound_exploration_obvious_winner(
    create_obvious_winner_matrix,
):
    """ """
    X, i = create_obvious_winner_matrix(10, 10)
    batch = upper_confidence_bound_exploration(X, a=1, batch_size=1)
    assert batch[0][0].item() == i


def test_upper_confidence_bound_exploration_nearly_full(create_nearly_full_matrix):
    """ """
    X, i = create_nearly_full_matrix(10, 10)
    batch = upper_confidence_bound_exploration(X, a=1, batch_size=1)
    assert batch[0][0].item() == i


def test_upper_confidence_bound_exploration_low_rank_factorization_obvious_winner(
    create_obvious_winner_matrix,
):
    """ """
    X, i = create_obvious_winner_matrix(10, 10)
    batch = upper_confidence_bound_exploration_low_rank_factorization(
        X, a=1, batch_size=1
    )
    assert batch[0][0].item() == i


def test_upper_confidence_bound_exploration_low_rank_factorization_nearly_full(
    create_nearly_full_matrix,
):
    """ """
    X, i = create_nearly_full_matrix(10, 10)
    batch = upper_confidence_bound_exploration_low_rank_factorization(
        X, a=1, batch_size=1
    )
    assert batch[0][0].item() == i



from typing import Any, Callable, Dict, Sequence

import torch
from tqdm.auto import tqdm

from banditeval.bandits import (
    upper_confidence_bound_exploration,
    upper_confidence_bound_exploration_low_rank_factorization,
)


class BanditEval:
    def __init__(
        self,
        methods: Sequence[Callable],
        dataset: Any,
        evaluation_function: Callable,
        algorithm: str,
        observation_matrix: torch.Tensor = None,
    ) -> None:
        if algorithm == "ucbe":
            self.step = upper_confidence_bound_exploration
        elif algorithm == "ucbe-lrf":
            self.step = upper_confidence_bound_exploration_low_rank_factorization
        else:
            raise ValueError(f"Unknown algorithm: {algorithm}.")
        self.observation_matrix = (
            torch.full((len(methods), len(dataset)), torch.nan)
            if observation_matrix is None
            else observation_matrix
        )

        self.methods = methods
        self.dataset = dataset
        self.evaluation_function = evaluation_function
        self.algorithm = algorithm

    def __call__(self, budget: int, **kwargs: Dict[Any, str]) -> torch.Tensor:
        for _ in tqdm(range(budget)):
            indices = self.step(self.observation_matrix, return_mus=False, **kwargs)
            if indices is None:
                break
            else:
                row_indices, column_indices = indices
            i = row_indices[0]
            method = self.methods[i]
            examples = [self.dataset[j.item()] for j in column_indices]
            scores = self.evaluation_function(method, examples)
            self.observation_matrix[row_indices, column_indices] = scores
        _, method_scores = self.step(self.observation_matrix, return_mus=True, **kwargs)
        return method_scores
